@@ -1,7 +1,9 @@
+from base64 import b64encode
 from datetime import datetime, timezone
 from email import utils
 from re import fullmatch
 from typing import Dict, List, Optional, Union
+from urllib.parse import quote as url_quote
 
 from kiss_headers.models import Header
 from kiss_headers.utils import class_to_header_name, prettify_header_name, quote
@@ -23,6 +25,10 @@ class CustomHeader(Header):
     __squash__: bool = False  # This value indicate whenever the representation of multiple entries should be squashed into one content.
     __tags__: List[str] = []
 
+    __override__: Optional[
+        str
+    ] = None  # Create this static member in your custom header when the class name does not match the target header.
+
     def __init__(self, initial_content: str = "", **kwargs: Optional[str]):
         """
         :param initial_content: Initial content of the Header if any.
@@ -33,7 +39,12 @@ class CustomHeader(Header):
                 "You can not instantiate CustomHeader class. You may create first your class that inherit it."
             )
 
-        super().__init__(class_to_header_name(self.__class__), initial_content)
+        super().__init__(
+            class_to_header_name(self.__class__)
+            if not self.__class__.__override__
+            else prettify_header_name(self.__class__.__override__),
+            initial_content,
+        )
 
         for attribute, value in kwargs.items():
             if value is None:
@@ -58,6 +69,11 @@ class Accept(CustomHeader):
         :param mime: Describe the MIME using this syntax <MIME_type/MIME_subtype>
         :param qualifier: Any value used is placed in an order of preference expressed using relative quality value called the weight.
         :param kwargs:
+        >>> header = Accept("text/html", qualifier=0.8)
+        >>> header.content
+        'text/html; q="0.8"'
+        >>> repr(header)
+        'Accept: text/html; q="0.8"'
         """
         if len(mime.split("/")) != 2:
             raise ValueError(
@@ -72,6 +88,17 @@ class Accept(CustomHeader):
             mime, **args,
         )
 
+    def get_mime(self) -> Optional[str]:
+        """Return defined mime in current accept header."""
+        for el in self.attrs:
+            if "/" in el:
+                return el
+        return None
+
+    def get_qualifier(self, _default: float = 1.0) -> Optional[float]:
+        """Return defined qualifier for specified mime. If not set, output 1.0."""
+        return float(str(self["q"])) if self.has("q") else _default
+
 
 class ContentType(CustomHeader):
     """
@@ -98,6 +125,9 @@ class ContentType(CustomHeader):
         :param format_: Mostly used in IMAP, could be one of : original or flowed.
         :param boundary: For multipart entities the boundary directive is required, which consists of 1 to 70 characters from a set of characters known to be very robust through email gateways, and not ending with white space. It is used to encapsulate the boundaries of the multiple parts of the message.
         :param kwargs:
+        >>> header = ContentType("text/html", charset="utf-8")
+        >>> repr(header)
+        'Content-Type: text/html; charset="UTF-8"'
         """
 
         if len(mime.split("/")) != 2:
@@ -115,6 +145,17 @@ class ContentType(CustomHeader):
 
         super().__init__(mime, **args)
 
+    def get_mime(self) -> Optional[str]:
+        """Return defined mime in content type."""
+        for el in self.attrs:
+            if "/" in el:
+                return el
+        return None
+
+    def get_charset(self, _default: Optional[str] = "ISO-8859-1") -> Optional[str]:
+        """Extract defined charset, if not present will return 'ISO-8859-1' by default."""
+        return str(self["charset"]) if self.has("charset") else _default
+
 
 class XContentTypeOptions(CustomHeader):
     """
@@ -130,6 +171,9 @@ class XContentTypeOptions(CustomHeader):
         """
         :param nosniff: see https://fetch.spec.whatwg.org/#x-content-type-options-header
         :param kwargs:
+        >>> header = XContentTypeOptions(nosniff=True)
+        >>> repr(header)
+        'X-Content-Type-Options: nosniff'
         """
         super().__init__("nosniff" if nosniff else "", **kwargs)
 
@@ -155,8 +199,8 @@ class ContentDisposition(CustomHeader):
         """
         :param disposition: Could be either inline, form-data, attachment or empty. Choose one. Default to inline.
         :param name: Is a string containing the name of the HTML field in the form that the content of this subpart refers to.
-        :param filename: Is a string containing the original name of the file transmitted. The filename is always optional and must not be used blindly by the application.
-        :param fallback_filename: Fallback filename if filename parameter does not uses the encoding defined in RFC 5987. ASCII-US Only.
+        :param filename: Is a string containing the original name of the file transmitted. The filename is always optional and must not be used blindly by the application. ASCII-US Only.
+        :param fallback_filename: Fallback filename if filename parameter does not uses the encoding defined in RFC 5987.
         :param boundary: For multipart entities the boundary directive is required, which consists of 1 to 70 characters from a set of characters known to be very robust through email gateways, and not ending with white space. It is used to encapsulate the boundaries of the multiple parts of the message.
         :param kwargs:
         """
@@ -170,7 +214,7 @@ class ContentDisposition(CustomHeader):
                 fallback_filename.encode("ASCII")
             except UnicodeEncodeError:
                 raise ValueError(
-                    "The fallback filename should only contain valid ASCII characters. Not '{fb_filename}'.".format(
+                    "The fallback filename should only contain valid ASCII characters. Not '{fb_filename}'. Use fallback_filename instead.".format(
                         fb_filename=fallback_filename
                     )
                 )
@@ -178,7 +222,9 @@ class ContentDisposition(CustomHeader):
         args: Dict = {
             "name": name,
             "filename": filename,
-            "filename*": fallback_filename,
+            "filename*": ("UTF-8''" + url_quote(fallback_filename))
+            if fallback_filename
+            else None,
             "boundary": boundary,
         }
 
@@ -198,10 +244,13 @@ class Authorization(CustomHeader):
 
     __tags__: List[str] = ["request"]
 
-    def __init__(self, type_: str, credentials: str):
+    def __init__(self, type_: str, credentials: str, **kwargs: Optional[str]):
         """
         :param type_: Authentication type. A common type is "Basic". See IANA registry of Authentication schemes for others.
         :param credentials: Associated credentials to use. Preferably Base-64 encoded.
+        >>> header = Authorization("Bearer", "base64encoded")
+        >>> repr(header)
+        'Authorization: Bearer base64encoded'
         """
         if type_.lower() not in [
             "basic",
@@ -222,10 +271,40 @@ class Authorization(CustomHeader):
             )
 
         super().__init__(
-            "{type_} {credentials}".format(type_=type_, credentials=credentials)
+            "{type_} {credentials}".format(type_=type_, credentials=credentials),
+            **kwargs,
         )
 
 
+class BasicAuthorization(Authorization):
+    """
+    Same as Authorization header but simplified for the Basic method. Also an example of __override__ usage.
+    """
+
+    __override__ = "Authorization"
+
+    def __init__(
+        self,
+        username: str,
+        password: str,
+        charset: str = "latin1",
+        **kwargs: Optional[str],
+    ):
+        """
+        :param username:
+        :param password:
+        :param kwargs:
+        >>> header = BasicAuthorization("azerty", "qwerty")
+        >>> header
+        Authorization: Basic YXplcnR5OnF3ZXJ0eQ==
+        """
+        b64_auth_content: str = b64encode(
+            (username + ":" + password).encode(charset)
+        ).decode("ascii")
+
+        super().__init__("Basic", b64_auth_content, **kwargs)
+
+
 class ProxyAuthorization(Authorization):
     """
     The HTTP Proxy-Authorization request header contains the credentials to authenticate a user agent to a proxy server,
@@ -255,6 +334,12 @@ class Host(CustomHeader):
         """
         :param host: The domain name of the server (for virtual hosting).
         :param port: TCP port number on which the server is listening.
+        >>> header = Host("www.python.org")
+        >>> repr(header)
+        'Host: www.python.org'
+        >>> header = Host("www.python.org", port=8000)
+        >>> repr(header)
+        'Host: www.python.org:8000'
         """
         super().__init__(host + (":" + str(port) if port else ""), **kwargs)
 
@@ -308,6 +393,10 @@ class Date(CustomHeader):
             **kwargs,
         )
 
+    def get_datetime(self) -> datetime:
+        """Parse and return a datetime according to content."""
+        return utils.parsedate_to_datetime(str(self))
+
 
 class CrossOriginResourcePolicy(CustomHeader):
     """
@@ -344,6 +433,9 @@ class Allow(CustomHeader):
         """
         :param supported_verb: Choose exactly one of "HEAD", "GET", "POST", "PUT", "PATCH", "DELETE", "PURGE", "CONNECT" or "TRACE" HTTP verbs.
         :param kwargs:
+        >>> header = Allow("POST")
+        >>> repr(header)
+        'Allow: POST'
         """
         supported_verb = supported_verb.upper()
 
@@ -384,6 +476,29 @@ class Digest(CustomHeader):
         super().__init__("", **args)
 
 
+class Cookie(CustomHeader):
+    """The Cookie HTTP request header contains stored HTTP cookies previously sent by
+    the server with the Set-Cookie header."""
+
+    __tags__ = ["request"]
+
+    def __init__(self, **kwargs: Optional[str]):
+        """
+        :param kwargs: Pair of cookie name associated with a value.
+        """
+        super().__init__("", **kwargs)
+
+    def get_cookies_names(self) -> List[str]:
+        """Retrieve all defined cookie names from Cookie header."""
+        return self.attrs
+
+    def get_cookie_value(
+        self, cookie_name: str, __default: Optional[str] = None
+    ) -> Optional[str]:
+        """Retrieve associated value with a given cookie name."""
+        return str(self[cookie_name]) if cookie_name in self else __default
+
+
 class SetCookie(CustomHeader):
     """
     The Set-Cookie HTTP response header is used to send cookies from the server to the user agent,
@@ -451,6 +566,34 @@ class SetCookie(CustomHeader):
         if is_httponly:
             self += "HttpOnly"  # type: ignore
 
+    def is_http_only(self) -> bool:
+        """Determine if the cookie can only be accessed by the browser."""
+        return "HttpOnly" in self
+
+    def is_secure(self) -> bool:
+        """Determine if the cookie is TLS/SSL only."""
+        return "Secure" in self
+
+    def get_expire(self) -> Optional[datetime]:
+        """Retrieve the parsed expiration date."""
+        return (
+            utils.parsedate_to_datetime(str(self["expires"]))
+            if self.has("expires")
+            else None
+        )
+
+    def get_max_age(self) -> Optional[int]:
+        """Getting the max-age value as an integer if set."""
+        return int(str(self["max-age"])) if "max-age" in self else None
+
+    def get_cookie_name(self) -> str:
+        """Extract the cookie name."""
+        return self.attrs[0]
+
+    def get_cookie_value(self) -> str:
+        """Extract the cookie value."""
+        return str(self[self.get_cookie_name()])
+
 
 class StrictTransportSecurity(CustomHeader):
     """
@@ -876,21 +1019,30 @@ class WwwAuthenticate(CustomHeader):
     """
     The HTTP WWW-Authenticate response header defines the authentication
     method that should be used to gain access to a resource.
+    Fair-Warning : This header is like none other and is harder to parse. It need a specific case.
     """
 
     __tags__: List[str] = ["response"]
+    __squash__ = True
 
     def __init__(
         self,
-        auth_type: str,
-        realm: str,
-        charset: Optional[str] = None,
+        auth_type: Optional[str] = None,
+        challenge: str = "realm",
+        value: str = "Secured area",
         **kwargs: Optional[str],
     ):
-        args: Dict = {"realm": realm, charset: charset.upper() if charset else None}
-        args.update(kwargs)
-
-        super().__init__(auth_type)
+        """
+        >>> www_authenticate = WwwAuthenticate("Basic", "realm", "Secured area")
+        >>> repr(www_authenticate)
+        'Www-Authenticate: Basic realm="Secured area"'
+        >>> headers = www_authenticate + WwwAuthenticate(challenge="charset", value="UTF-8")
+        >>> repr(headers)
+        'Www-Authenticate: Basic realm="Secured area", charset="UTF-8"'
+        """
+        super().__init__(
+            f'{auth_type+" " if auth_type else ""}{challenge}="{value}"', **kwargs
+        )
 
 
 class XDnsPrefetchControl(CustomHeader):
