@@ -15,35 +15,18 @@ from typing import (
 
 from kiss_headers.structures import CaseInsensitiveDict
 from kiss_headers.utils import (
+    extract_comments,
     header_content_split,
     header_name_to_class,
     header_strip,
     is_legal_header_name,
     normalize_str,
     prettify_header_name,
+    unfold,
+    unpack_protected_keyword,
     unquote,
 )
 
-RESERVED_KEYWORD: List[str] = [
-    "and_",
-    "assert_",
-    "in_",
-    "not_",
-    "pass_",
-    "finally_",
-    "while_",
-    "yield_",
-    "is_",
-    "as_",
-    "break_",
-    "return_",
-    "elif_",
-    "except_",
-    "def_",
-    "from_",
-    "for_",
-]
-
 OUTPUT_LOCK_TYPE: bool = False
 
 
@@ -80,15 +63,12 @@ class Header(object):
         self._pretty_name: str = prettify_header_name(self._name)
         self._content: str = content
 
-        self._members: List[str] = [
-            el.lstrip() for el in header_content_split(self._content, ";")
-        ]
+        self._members: List[str] = header_content_split(self._content, ";")
 
         self._not_valued_attrs: List[str] = list()
-        self._valued_attrs: MutableMapping[str, Union[str, List[str]]] = dict()
-        self._valued_attrs_normalized: MutableMapping[
+        self._valued_attrs: MutableMapping[
             str, Union[str, List[str]]
-        ] = dict()
+        ] = CaseInsensitiveDict()
 
         for member in self._members:
             if member == "":
@@ -110,9 +90,6 @@ class Header(object):
                     else:
                         self._valued_attrs[key].append(value)  # type: ignore
 
-                self._valued_attrs_normalized[normalize_str(key)] = self._valued_attrs[
-                    key
-                ]
                 continue
 
             self._not_valued_attrs.append(unquote(member))
@@ -152,11 +129,60 @@ class Header(object):
 
         return self._content
 
+    @property
+    def unfolded_content(self) -> str:
+        """Output unfolded associated content to header. Meaning that every LF + n space(s) would be properly
+        replaced."""
+        return unfold(self.content)
+
+    @property
+    def comments(self) -> List[str]:
+        """Retrieve comments in header content."""
+        return extract_comments(self.content)
+
+    def __lt__(self, other: object) -> bool:
+        """
+        This method is only implemented to make sorted work with Header.
+        The lower than is based on alphabetical order using the header name.
+        >>> Header("A", "") < Header("Z", "")
+        True
+        """
+        if not isinstance(other, Header):
+            raise NotImplementedError
+        return self.normalized_name < other.normalized_name
+
+    def __gt__(self, other: object) -> bool:
+        """
+        This method is only implemented to make sorted work with Header.
+        The greater than is based on alphabetical order using the header name.
+        >>> Header("A", "") > Header("Z", "")
+        False
+        """
+        if not isinstance(other, Header):
+            raise NotImplementedError
+        return self.normalized_name > other.normalized_name
+
     def __deepcopy__(self, memodict: Dict) -> "Header":
         """Simply provide a deepcopy of an Header object. Pointer/Reference free of the initial reference."""
         return Header(deepcopy(self.name), deepcopy(self.content))
 
     def __iadd__(self, other: Union[str, "Header"]) -> "Header":
+        """
+        Allow you to assign-add any string to an Header instance. The string will be a new member of your header.
+        >>> header = Header("X-Hello-World", "")
+        >>> repr(header)
+        'X-Hello-World: '
+        >>> header += "preload"
+        >>> repr(header)
+        'X-Hello-World: preload'
+        >>> header += "inclSubDomain"
+        >>> repr(header)
+        'X-Hello-World: preload; inclSubDomain'
+        >>> header += 1
+        Traceback (most recent call last):
+          ...
+        TypeError: Cannot assign-add with type <class 'int'> to an Header.
+        """
         if not isinstance(other, str):
             raise TypeError(
                 "Cannot assign-add with type {type_} to an Header.".format(
@@ -171,6 +197,21 @@ class Header(object):
         return self
 
     def __add__(self, other: Union[str, "Header"]) -> Union["Header", "Headers"]:
+        """
+        This implementation permit to add either a string or a Header to your Header instance.
+        When you add string to your Header instance, it will create another instance with a new
+        member in it using the string; see iadd doc about it. But when its another Header the result is an Headers
+        object containing both Header object.
+        >>> headers = Header("X-Hello-World", "1") + Header("Content-Type", "happiness=True")
+        >>> len(headers)
+        2
+        >>> headers.keys()
+        ['X-Hello-World', 'Content-Type']
+        >>> Header("Content-Type", "happiness=True") + 1
+        Traceback (most recent call last):
+          ...
+        TypeError: Cannot make addition with type <class 'int'> to an Header.
+        """
         if not isinstance(other, str) and not isinstance(other, Header):
             raise TypeError(
                 "Cannot make addition with type {type_} to an Header.".format(
@@ -211,7 +252,7 @@ class Header(object):
 
         other = normalize_str(other)
 
-        if other in self._valued_attrs_normalized:
+        if other in self._valued_attrs:
             del self[other]
 
         if other in self._not_valued_attrs:
@@ -245,23 +286,19 @@ class Header(object):
         """
 
         # Avoid conflict with __init__ sequence of Header
-        if key in [
+        if key in {
             "_name",
             "_normalized_name",
             "_pretty_name",
             "_content",
             "_members",
             "_not_valued_attrs",
-            "_valued_attrs_normalized",
             "_valued_attrs",
-        ]:
+            "__class__",
+        }:
             return super().__setattr__(key, value)
 
-        if key[0] == "_":
-            key = key[1:]
-
-        if key.lower() in RESERVED_KEYWORD:
-            key = key[:-1]
+        key = unpack_protected_keyword(key)
 
         self[key] = value
 
@@ -270,7 +307,6 @@ class Header(object):
         Set an attribute bracket syntax like. This will erase previously set attribute named after the key.
         Any value that are not a str are casted to str.
         """
-        key_normalized = normalize_str(key)
 
         if key in self:
             del self[key]
@@ -278,7 +314,6 @@ class Header(object):
             value = str(value)
 
         self._valued_attrs[key] = value
-        self._valued_attrs_normalized[key_normalized] = self._valued_attrs[key]
 
         self._content += '{semi_colon_r}{key}="{value}"'.format(
             key=key,
@@ -296,22 +331,14 @@ class Header(object):
         >>> str(headers.content_type)
         'text/html'
         """
-        key_normalized = normalize_str(key)
-
-        if key_normalized not in self._valued_attrs_normalized:
+        if key not in self._valued_attrs:
             raise KeyError(
                 "'{item}' attribute is not defined within '{header}' header.".format(
                     item=key, header=self.name
                 )
             )
 
-        del self._valued_attrs_normalized[key]
-        not_normalized_keys = self._valued_attrs.keys()
-
-        for key_ in not_normalized_keys:
-            if normalize_str(key_) == key_normalized:
-                del self._valued_attrs[key_]
-                break
+        del self._valued_attrs[key]
 
         for elem in findall(
             r"{key_name}=.*?(?=[;\n])".format(key_name=escape(key)),
@@ -332,7 +359,7 @@ class Header(object):
         """
         item = normalize_str(item)
 
-        if item not in self._valued_attrs_normalized:
+        if item not in self._valued_attrs:
             raise AttributeError(
                 "'{item}' attribute is not defined within '{header}' header.".format(
                     item=item, header=self.name
@@ -400,7 +427,7 @@ class Header(object):
         Provide a better auto-completion when using python interpreter. We are feeding __dir__ so Python can be aware
         of what properties are callable. In other word, more precise auto-completion when not using IDE.
         """
-        return list(super().__dir__()) + list(self._valued_attrs_normalized.keys())
+        return list(super().__dir__()) + list(self._valued_attrs.keys())
 
     @property
     def attrs(self) -> List[str]:
@@ -420,21 +447,47 @@ class Header(object):
     def get(self, attr: str) -> Optional[Union[str, List[str]]]:
         """
         Retrieve associated value of an attribute.
+        >>> header = Header("Content-Type", "application/json; charset=UTF-8; format=flowed")
+        >>> header.charset
+        'UTF-8'
+        >>> header.ChArSeT
+        'UTF-8'
+        >>> header.FORMAT
+        'flowed'
+        >>> header.format
+        'flowed'
         """
         if attr not in self._valued_attrs:
             return None
         return self._valued_attrs[attr]
 
-    def __getitem__(self, item: Union[str]) -> Union[str, List[str]]:
+    def has_many(self, name: str) -> bool:
         """
-        This method will allow you to retrieve attribute value using the bracket syntax, list-like.
+        Determine if an attribute name has multiple entries in Header. Detect OneToMany entries.
+        >>> header = Header("A", "charset=UTF-8; charset=ASCII; format=flowed")
+        >>> header.has_many("charset")
+        True
+        >>> header.has_many("format")
+        False
         """
-        normalized_item = normalize_str(item)
+        if name not in self:
+            return False
+
+        r = self[name]
+
+        return isinstance(r, list) and len(r) > 1
+
+    def __getitem__(self, item: Union[str, int]) -> Union[str, List[str]]:
+        """
+        This method will allow you to retrieve attribute value using the bracket syntax, list-like or dict-like.
+        """
+        if isinstance(item, int):
+            return (
+                self._members[item] if not OUTPUT_LOCK_TYPE else [self._members[item]]
+            )
 
         if item in self._valued_attrs:
             value = self._valued_attrs[item]
-        elif normalized_item in self._valued_attrs_normalized:
-            value = self._valued_attrs_normalized[normalized_item]
         else:
             raise KeyError(
                 "'{item}' attribute is not defined within '{header}' header.".format(
@@ -446,9 +499,9 @@ class Header(object):
             value = [value]
 
         return (
-            unquote(value)
+            unfold(unquote(value))
             if not isinstance(value, list)
-            else [unquote(v) for v in value]
+            else [unfold(unquote(v)) for v in value]
         )
 
     def __getattr__(self, item: str) -> Union[str, List[str]]:
@@ -456,16 +509,9 @@ class Header(object):
         All the magic happen here, this method should be invoked when trying to call (not declared) properties.
         For instance, calling self.charset should end up here and be replaced by self['charset'].
         """
-        if item[0] == "_":
-            item = item[1:]
+        item = unpack_protected_keyword(item)
 
-        if item.lower() in RESERVED_KEYWORD:
-            item = item[:-1]
-
-        if (
-            item not in self._valued_attrs
-            and normalize_str(item) not in self._valued_attrs_normalized
-        ):
+        if item not in self._valued_attrs:
             raise AttributeError(
                 "'{item}' attribute is not defined within '{header}' header.".format(
                     item=item, header=self.name
@@ -483,7 +529,7 @@ class Header(object):
         item = normalize_str(item)
         for attr in self.attrs:
             target = normalize_str(attr)
-            if item == target or item in target.split(","):
+            if item == target or item in header_content_split(target, " "):
                 return True
         return False
 
@@ -492,6 +538,7 @@ class Headers(object):
     """
     Object oriented representation for Headers. Contains a list of Header with some level of abstraction.
     Combine advantages of dict, CaseInsensibleDict and native objects.
+    Headers does not inherit of the Mapping type, but it does borrow some concept from it.
     """
 
     # Most common headers that you may or may not find. This should be appreciated when having auto-completion.
@@ -552,11 +599,15 @@ class Headers(object):
 
     from_: Union[Header, List[Header]]
 
-    def __init__(self, headers: Optional[List[Header]] = None):
+    def __init__(self, *headers: Union[List[Header], Header]):
         """
         :param headers: Initial list of header. Can be empty.
         """
-        self._headers: List[Header] = headers or []
+        self._headers: List[Header] = headers[0] if len(headers) == 1 and isinstance(
+            headers[0], list
+        ) else list(
+            headers  # type: ignore
+        )
 
     def has(self, header: str) -> bool:
         """
@@ -572,6 +623,22 @@ class Headers(object):
             return None
         return self[header]
 
+    def has_many(self, name: str) -> bool:
+        """
+        Determine if an header name has multiple entries in Headers. Detect OneToMany entries.
+        >>> headers = Header("A", "0") + Header("A", "1") + Header("B", "sad")
+        >>> headers.has_many("a")
+        True
+        >>> headers.has_many("b")
+        False
+        """
+        if name not in self:
+            return False
+
+        r = self[name]
+
+        return isinstance(r, list) and len(r) > 1
+
     def __iter__(self) -> Iterator[Header]:
         """
         Act like a list by yielding one element at a time. Each element is a Header object.
@@ -582,6 +649,7 @@ class Headers(object):
     def keys(self) -> List[str]:
         """
         Return a list of distinct header name set in headers.
+        Be aware that it wont return a typing.KeysView
         """
         keys = list()
 
@@ -592,9 +660,21 @@ class Headers(object):
 
         return keys
 
+    def values(self) -> NotImplemented:
+        """
+        I choose not to implement values() on Headers as it would bring more confusion..
+        Either we make it the same len as keys() or we don't. Either way don't please me. Hope to ear from the
+        community about this.
+        """
+        return NotImplemented
+
     def items(self) -> List[Tuple[str, str]]:
         """
         Provide an iterator witch each entry contain a tuple of header name and content.
+        This wont return a ItemView.
+        >>> headers = Header("X-Hello-World", "1") + Header("Content-Type", "happiness=True") + Header("Content-Type", "happiness=False")
+        >>> headers.items()
+        [('X-Hello-World', '1'), ('Content-Type', 'happiness=True'), ('Content-Type', 'happiness=False')]
         """
         items: List[Tuple[str, str]] = []
 
@@ -861,15 +941,13 @@ class Headers(object):
         Where the magic happen, every header are accessible via the property notation.
         The result is either a single Header or a list of Header.
         eg.
-        >>> headers = Header("Content-Type", "text/html; charset=UTF-8") + Header("Allow", "POST")
+        >>> headers = Header("Content-Type", "text/html; charset=UTF-8") + Header("Allow", "POST") + Header("From", "john-doe@gmail.com")
         >>> headers.content_type
         Content-Type: text/html; charset=UTF-8
+        >>> headers.from_
+        From: john-doe@gmail.com
         """
-        if item[0] == "_":
-            item = item[1:]
-
-        if item.lower() in RESERVED_KEYWORD:
-            item = item[:-1]
+        item = unpack_protected_keyword(item)
 
         if item not in self:
             raise AttributeError(
@@ -894,6 +972,17 @@ class Headers(object):
         """
         return repr(self).encode("utf-8", errors="surrogateescape")
 
+    def __reversed__(self) -> "Headers":
+        """Return a new instance of Headers containing headers in reversed order."""
+        list_of_headers: List[Header] = deepcopy(self._headers)
+        list_of_headers.reverse()
+
+        return Headers(list_of_headers)
+
+    def __bool__(self) -> bool:
+        """Return True if Headers does contain at least one entry in it."""
+        return bool(self._headers)
+
     def __contains__(self, item: Union[Header, str]) -> bool:
         """
         This method will allow you to test if a header, based on it's string name, is present or not in headers.
@@ -909,6 +998,29 @@ class Headers(object):
 
         return False
 
+    def pop(self, __index_or_name: Union[str, int] = -1) -> Union[Header, List[Header]]:
+        """Pop header from headers. By default the last one."""
+        if isinstance(__index_or_name, int):
+            return self._headers.pop(__index_or_name)
+        if isinstance(__index_or_name, str):
+            headers = self.get(__index_or_name)
+
+            if headers is None:
+                raise IndexError()
+
+            if isinstance(headers, list):
+                for header in headers:
+                    self._headers.remove(header)
+            else:
+                self._headers.remove(headers)
+            return headers
+        raise TypeError
+
+    def popitem(self) -> Tuple[str, str]:
+        """Pop last header as a tuple (header name, header content)."""
+        header: Header = self.pop()  # type: ignore
+        return header.name, header.content
+
     def __dir__(self) -> Iterable[str]:
         """
         Provide a better auto-completion when using python interpreter. We are feeding __dir__ so Python can be aware
